function [ConfusionMatrix] = determine_confusion_matrix( distNeuronsToData, Labels )
%
% Determines the confusion matrix
%
% A.Hasenfuss (c) 2008
%
NeuronLabels = label_neurons( distNeuronsToData, Labels );
[cLabels] = label_data( distNeuronsToData, NeuronLabels );
ConfusionMatrix = determine_confusion_matrix_given_labels( cLabels, Labels );    

function [Assignment] = assign_datapoints_to_neurons( distNeuronsToData )

%
% Assign datapoints to nearest neuron
% 
%
% A.Hasenfuss (c) 2005
%

[dummy,idx] = min(distNeuronsToData,[],1);
Assignment = idx';

function [NeuronLabels, Classification, Assignment, Hitrate] = label_neurons( distNeuronsToData, Labels )
%
% Alexander Hasenfuss (c) 2005-2007
%
% Labels the neurons according to the majority-label of the data points within their
% receptive fields.
%
% It also stores the (fuzzy) classification for each neuron, the assignment of
% data to neurons, and the hitrate.
%
% Alternative to function "label_neurons_euclidean"
%
% Returns
% NeuronLabels ... (n x 1) Matrix 
% Classification ... {n x 1} Cell Array containing (k x 3) matrices with output of tabulate-command
% Assignment ... (m x 1) Matrix
%

% Local variables
sum_blacksheep = 0;

% Needed values
[num_neurons,num_data] = size( distNeuronsToData );

% Return values
NeuronLabels = zeros(num_neurons,1);
Classification = cell(num_neurons,1);

% Determine the corresponding neurons for all datapoints
[Assignment] = assign_datapoints_to_neurons( distNeuronsToData );

% For each neuron...
for i = 1:num_neurons,
    
    % ...determine the indexes of the corresponding datapoints
    idx = find( Assignment == i );
    
    if numel( idx ) ~= 0,
            
        % Get the labels of the corresponding datapoints (receptive field)
        corresp_labels = Labels( idx );
    
        % Count occurrence of label...
        tab = tabulate( corresp_labels );
    
        % ...and declare a winner label
        [dummy, max_idx] = max( tab(:,2) );
        winner_label = tab(max_idx,1);
        
        % Sum up black sheep
        sum_blacksheep = sum_blacksheep + (sum(tab(:,2)) - tab(max_idx,2));
    
        % Store represented labels and remove zero lines before
        Classification{i,1} = tab(find(tab(:,2) ~= 0),:);

    else
        winner_label = NaN;
    end;
    
    % Store winner label for this neuron
    NeuronLabels(i) = winner_label;

end;

Hitrate = 1 - (sum_blacksheep / num_data);

function [cLabels] = label_data( distNeuronsToData, NeuronLabels )
%
% Alexander Hasenfuss (c) 2005-2007
%
% Labels the data according to the corresponding winner neuron
%

% Determine winner neurons
[dummy,idx] = min( distNeuronsToData, [], 1 );

% Use assignment as index to neuron labels 
cLabels = NeuronLabels( idx );

function ConfusionMatrix = determine_confusion_matrix_given_labels( cLabels, Labels )

%
% Determines the confusion matrix
%
% Each entry (i,j) of the resulting matrix corresponds to
% the number of items from class i that were wrongly classified
% to belong to class j.
%
% Param: cLabels ... calculated labels
%        Labels  ... "true" labels
%
% A. Hasenfuss (c) 2007
%
% see also determine_confusion_matrix(distNeuronsToData,Labels) !!
%

% TODO: Check parameters! size(cLabels) == size(Labels), etc...
%       Change parameter order!!
 
% Check and correct labels
[xL,yL] = size( Labels );
if xL < yL
  Labels = reshape(Labels, yL, xL);
end

% Get class information
cT = tabulate( cLabels );
T = tabulate( Labels );
num_T = size(T,1);
num_cT = size(cT,1);

% Storage
if num_T < num_cT, error('More cLabel-classes than Label-classes!'); end
ConfusionMatrix = zeros(num_T, num_T);
idxT = cell(1,num_T);

% Separate datapoints according to true classes...
for i = 1:num_T
  idxT{i} = find( Labels == T(i,1) );
end
    
% For each calculated class j determine number of contained points of true class i...
for j = 1:num_cT

  % Determine assigned datapoints of calculated class j
  idxC = find( cLabels == cT(j,1) );
  
  % For each true-class...
  for i = 1:num_T
    
    % Count datapoints of true class i that are included in calculated class j
    its = intersect(  idxT{i}, idxC );
    ConfusionMatrix(i,j) = numel( its );
  
  end
    
end

