%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Relational Generalized Learning Vector Quantizer (R-GLVQ)
% With optimization based on Broyden-Fletcher-Goldfarb-Shanno 
% or other types of optimization schemes - see parameters
%
% - a classifier for relational data - 
% GLVQ (with positive updates only) for relational data
% (C) Frank-Michael Schleif (c) 2011-2012
% University of Bielefeld
% fschleif@techfak.uni-bielefeld.de
%
%
% Input:
% 	- mDistances 		(m x m)  matrix of dissimilarities (main diagonal should be zero)
%         vLabels    		(m x 1)  a numerical vector of class lables (0,0,0,1,1,1,...,2,2,2 ...)
%         optional              (struct) a parameter object in accordance to the function fGetDefaultParameter
%
% Output:
%        oModel			(struct) a struct with the updated mAlpha (n x m) matrix with
%                               n - number of prototypes; a label vector of the prototypes
%                               an accuracy estimate and a runtime value, further an assignment
%                               vector (n x 1) is given providing information of the cluster 
%                               assignments of each training point
%
% This is test code only - some features are off/on see conditions in the
% code, some others can be defined by the parameter object
%
% See sample code how to use this algorithm and how to apply the model to new data
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [oModel] = rglvq_train_neg(mDistances,vLabels,varargin)
    iTimer=cputime;
    if(1) % this will change the distances - ensure this outside
        mDistances = fCheckDissimilarity(mDistances); % check and try to fix if necessary
    end
    % get parameter set
    if(nargin<3)
        oParameter = fGetDefaultParameter();
    else
        oParameter = varargin{1};
    end

    % set local parameters 
    oModel        = [];
    [N,~]         = size(mDistances);
    
    iCycles       = oParameter.iCycles;
    bVisualize    = oParameter.bVisualize;
    vUniqueClasses= unique(vLabels);
    iNrClasses    = length(vUniqueClasses);
    iNrPrototypes = iNrClasses * oParameter.iPrototypes;
    iK            = oParameter.iK;
    % fix class labels    
    vCodeLabels   = reshape(repmat(vUniqueClasses',oParameter.iPrototypes,1),1, iNrPrototypes);    
    bGivenProtos  = oParameter.bGivenProtos;

    if(isfield(oParameter,'bNyStroem') && oParameter.bNyStroem == true)
        if(oParameter.iPNy < 0 || oParameter.iPNy > 1)
            oParameter.iPNy = 0.1;
        end  
        oNy              = CalculateNystroem(mDistances,oParameter.iPNy);
        oT               = oNy.W * oNy.Cp;
        Cp               = oNy.Cp;
        C                = oNy.C;
        Khat             = oNy.Cp(:,oNy.vLandMarks);        
        fHelper          = @(cNeurons)(oT*cNeurons')';        
        %fInnerDistPtr    = @(Temp,iSampleIndex)Temp * Cp(:,iSampleIndex)-0.5*diag(Temp * Khat * Temp');
        fInnerDistPtr    = @(Temp,iSampleIndex)Temp * Cp(:,iSampleIndex)-0.5*diag(oT' * Khat * oT)';
        %fInnerDistPtr2   = @(Temp)(bsxfun(@minus,Temp * Cp,0.5*diag(Temp * Khat * Temp')));
        fInnerDistPtr2   = @(Temp)(bsxfun(@minus,Temp * Cp,0.5*diag(oT' * Khat * oT)'));
  
        
        fHelperRange     = @(vNeuron,vRange)(oT(:,vRange)*vNeuron')';
        fDistPtrRange    = @(vNeuron,iSampleIndex,vRange)((fInnerDistPtr(fHelperRange(vNeuron,vRange),iSampleIndex)));        
        fDistPtr2        = @(x)((fInnerDistPtr2(fHelper(x))));
        
        fRHelperRange    = @(vNeuron,vRange)(oT(:,vRange)*vNeuron(:,vRange)')';
        %fRInnerDistPtr2  = @(Temp,vRange)(bsxfun(@minus,Temp * Cp(:,vRange),0.5*diag(Temp * Khat * Temp')));
        fRInnerDistPtr2  = @(Temp,vRange)(bsxfun(@minus,Temp * Cp(:,vRange),0.5*diag(oT' * Khat * oT)'));
        fRDistPtr2       = @(x,vRange)((fRInnerDistPtr2(fRHelperRange(x,vRange),vRange)));
        
        fDistParAlphaPtr2ndRange = @(Neuron,vRange)bsxfun(@minus,C(vRange,:) * oT,(Neuron*oT')*Cp)';
        
        fVisualizePtr    = @(x)fVisualize(x,fDistPtr2,oParameter,iNrPrototypes,N);
        oModel.vLandMarks= oNy.vLandMarks;
    else                
        fDistPtrRange    = @(vNeuron ,iSampleIndex,vRange)(vNeuron * mDistances(vRange,iSampleIndex) - 0.5 * sum( (vNeuron * mDistances(vRange,vRange)) .* vNeuron,2)); % shortcut if alpha is sparse        
        fHelper          = @(cNeurons)(cNeurons*mDistances);
        fInnerDistPtr2   = @(Temp,cNeurons)(bsxfun(@minus,Temp,0.5 * sum(bsxfun(@times,Temp,cNeurons),2)));
        fDistPtr2        = @(cNeurons)((fInnerDistPtr2(fHelper(cNeurons),cNeurons)));

        fRHelper          = @(cNeurons,vRange)(cNeurons*mDistances(:,vRange));
        %fRDistPtr2        = @(cNeurons,vRange)((fInnerDistPtr2(fRHelper(cNeurons,vRange),cNeurons(:,vRange))));
        fRDistPtr2        = @(cNeurons,vRange)max((fInnerDistPtr2(fRHelper(cNeurons,vRange),cNeurons)),realmin);
        
        %fDistParAlphaPtr2ndRange = @(Neuron,vRange)bsxfun(@minus,mDistances(:,vRange), (Neuron * mDistances(:,vRange)));
        fDistParAlphaPtr2ndRange = @(Neuron,vRange)bsxfun(@minus,mDistances(:,vRange)', (Neuron * mDistances))';
                
        fVisualizePtr    = @(x)fVisualize(x,fDistPtr2,oParameter,iNrPrototypes,N);
    end    

    if bGivenProtos
	    mAlpha      = oParameter.mInitProtos;
        vCodeLabels = oParameter.vInitProtoLabels;
        iNrPrototypes = numel(vCodeLabels);
    else
	    if(1) % random init 
            mAlpha        = rand(iNrPrototypes,N);
	    else  % class specific init - recommended - will also speed up later, because mAlpha gets sparse
            mAlpha        = rand(iNrPrototypes,N);
            iCounter      = 1;
            for(c=1:length(vUniqueClasses))
                vCurrentIndices      = (vLabels==vUniqueClasses(c));
                mAlpha(iCounter:iCounter+oParameter.iPrototypes-1,~vCurrentIndices) = realmin; % switch other entries off 
                iCounter = iCounter+oParameter.iPrototypes;
            end
	    end
    end
    mAlpha(mAlpha<0) = 1e-10;
    vOnes         = ones(1,N);    
    % Normalize
    Normalizer    = sum(mAlpha,2);
    Normalizer    = Normalizer(:,vOnes);
    mAlpha        = mAlpha./Normalizer;
    vError        = zeros(iCycles,1);
    [M,~]         = size(mAlpha);
    [N]           = length(vLabels);
    
    fErrorFctPtr         = @(mAlpha)fError(fDistPtr2,vCodeLabels,vLabels,mAlpha);
    ProgParClassInstance = ProgressParameterClass(fErrorFctPtr);
    ProgParClassInstance.N = N;
    ProgParClassInstance.iPrototypes = iNrPrototypes;
    LVQProgressPtr       = @(variables, output, where)RGLVQ_progresser(ProgParClassInstance,variables, output, where);
    LabelEqualsPrototype = bsxfun(@eq,vLabels,vCodeLabels);
    fCostFct             = @(variables)rglvq_optfun(variables,fRDistPtr2,fDistParAlphaPtr2ndRange,LabelEqualsPrototype);    

    
    %%% optimization options
    options = struct( ...
      'Display','off', ...
      'GradObj','on', ...
      'GradConstr',false, ...
      'GoalsExactAchieve',0, ...
      'TolFun',1e-6, ...
      'MaxIter',100, ...
      'MaxFunEvals', 100, ...
      'TolX',1e-6, ...
      'DiffMinChange',1e-6, ...
      'OutputFcn',LVQProgressPtr, ...      
      'HessUpdate','steepdesc' ...
    );      %'lbfgs', 'bfgs', 'steepdesc'
        
    opts.LS = 2;
    opts.LS_init = 1;
    opts.MaxIter = 10;
    opts.Method = 'sd';
    opts.Display = 'final';

    for(t=1:iCycles) % extra cycle optimization 
%        [mAlpha,fval]     = fminlbfgs(@(variables) fCostFct(variables), mAlpha, options);     
        mAlpha = minFunc(@(variables) fCostFct(variables), mAlpha(:),opts);        
        mAlpha = reshape(mAlpha, iNrPrototypes,N);
 
        % normalize
        mAlpha(mAlpha(:)<0) = 1e-10;
        mAlpha = bsxfun(@rdivide,mAlpha,sum(mAlpha,2));    
        
        if(0 & iK > 0)
            % if alphas are >= 0 - use standard k-approx        
            for(n=1:iNrPrototypes)
                [Y,I]  = sort(mAlpha(n,:),'descend');
                vAlphaBack = mAlpha(n,:);
                mAlpha(n,:) = 0;
                mAlpha(n,I(1:iK)) = vAlphaBack(I(1:iK));
                mAlpha(n,:) = mAlpha(n,:)./sum(mAlpha(n,:));
            end        
        end
        
        if(bVisualize && mod(t,1)==0)
            fVisualizePtr(mAlpha);            
            [~,vClusters] = min(fDistPtr2(mAlpha));
            dAcc          = sum(vCodeLabels(vClusters)==vLabels')./N
        end
        
        vError(t)  = rglvq_cost(mAlpha, fDistPtr2,LabelEqualsPrototype);
    end
    if(iK > 0)
        if(0)
            % if alphas are >= 0 - use standard k-approx        
            for(n=1:iNrPrototypes)
                [Y,I]  = sort(mAlpha(n,:),'descend');
                vAlphaBack = mAlpha(n,:);
                mAlpha(n,:) = 0;
                mAlpha(n,I(1:iK)) = vAlphaBack(I(1:iK));
                mAlpha(n,:) = mAlpha(n,:)./sum(mAlpha(n,:));
            end        
        else
            mAlpha = fKApprox(mAlpha,mDistances,iK,fDistPtrRange);
        end
    end
    
    [~,vClusters] = min(fDistPtr2(mAlpha));
    dAcc          = sum(vCodeLabels(vClusters)==vLabels')./N
    
    oModel.mAlpha       = mAlpha;
    distNeuronsToTrainData = fDistPtr2(mAlpha);
    [~,vClusters]       = min(distNeuronsToTrainData);
    oModel.vCodeLabels  = vCodeLabels;
    oModel.dAcc         = sum(vCodeLabels(vClusters)==vLabels')./N;
    oModel.vAssignments = vClusters;
    oModel.vError       = vError;
    oModel.iRunTime     = cputime-iTimer;
    oModel.distNeuronsToTrainData = distNeuronsToTrainData;
   
    
function fVisualize(mAlpha,fDist,oParameters,M,N)
    [~,vClusters]   = min(fDist(mAlpha));
    mMap            = zeros(M,N);
    mColor          = [0:1/length(vClusters):1];
    for(k=1:length(vClusters))
        iAssignedCluster = vClusters(k);        
        vEntryIndicesInCluster  = find(vClusters==iAssignedCluster);
        for(l=1:length(vEntryIndicesInCluster))
            i = vEntryIndicesInCluster(l);
            mMap(k,i) = mColor(iAssignedCluster);
            mMap(i,k) = mMap(k,i);
        end
    end
    if(~isfield(oParameters,'oData'))
        subplot(1,2,1), imagesc(mMap);    
        subplot(1,2,2), imagesc(mAlpha);
    else
        vPrototypes = mAlpha * oParameters.oData;
      %  subplot(1,3,1), imagesc(mMap);    
      %  subplot(1,3,2), imagesc(mAlpha);
      %  h=subplot(1,3,3);
      %  cla(h);
      %  hold;plot(vPrototypes(:,1),vPrototypes(:,2),'rx');plot(oParameters.oData(:,1),oParameters.oData(:,2),'go');hold off;
        h=subplot(1,1,1);
        cla(h);
        hold;plot(vPrototypes(:,3),vPrototypes(:,4),'rx');plot(oParameters.oData(:,3),oParameters.oData(:,4),'go');hold off;
        drawnow;
    end
    pause(1);

function [oNy] = CalculateNystroem(mDistances,iPNy)
    [N,~]              = size(mDistances);    
    iNumberOfLandmarks = ceil(N*iPNy);
    if(iNumberOfLandmarks == 0), 
        warning('To few landmarks - set to 1 only - result maybe useless');
        iNumberOfLandmarks = vRand(1); 
    end
    vLandMarks = randsample(N,iNumberOfLandmarks);
    C          = mDistances(:,vLandMarks);
    W          = pinv(C(vLandMarks,:));
    oNy        = struct('C',C,'W',W,'Cp',C','iPNy',iPNy,'vLandMarks',vLandMarks);

function [dAcc,dError] = fError(fDist,vCodeLabels,vDataLabels,mAlpha)
    mAlpha(mAlpha(:)<0) = 1e-10;
	mAlpha = bsxfun(@rdivide,mAlpha,sum(mAlpha,2));    
    mDist         = fDist(mAlpha); 
    mLabelMatch   = bsxfun(@eq,vCodeLabels,vDataLabels);
    dError        = 0;    
    for(i=1:length(vDataLabels))
        [dPError,~] = min(mDist((mLabelMatch(i,:)),i));
        [dMError,~] = min(mDist(~(mLabelMatch(i,:)),i));
        dError  = dError + (dPError - dMError) / (dPError+dMError);
    end
    [~,vClusters] = min(mDist);
	dAcc          = sum(vCodeLabels(vClusters)==vDataLabels')./numel(vDataLabels);

function [oParameter] = fGetDefaultParameter
    oParameter = struct('iCycles',100,'iPrototypes',3,'iK',0,'bVisualize',true,'bWithNeighborhood',false);
